!*****************************************
! multistep.f90 - timestepping subroutines
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

subroutine multistep
implicit none

include 'globalpm.fh'
include 'mpif.h'
 
real(kind=4) :: da,ra,dt_e,astemp
integer(kind=4) :: i,ierr,j,k
character(len=MSL) :: tsfile,hlfile

! advance timestep

nts=nts+1

! copy dt from previous timestep

if (nts.ne.1) dt_old=dt

! if the simulation was restarted, or if a checkpoint/projections
! were written to disk in the last timestep, then the final half-step
! update was already done

if (restart_step.or.check_step) then
    dt_old=0.0
    check_step=.false.
endif

if (proj_step) then
  dt_old=0.0
  proj_step=.false.
endif

! only the master process calculates timestep variables, after which
! they are broadcast to the other nodes

if (master) then

! if we are omitting the coarse grid - useful for testing purposes

  if (TS_RATIO_MAX.eq.0) then

    dt_e=0.5

    do
      call scale_factor(a,da,dt_e)
      ra=da/(a+da)
      if (ra.gt.RA_MAX) then
        dt_e=dt_e*(RA_MAX/ra)
      else
        exit
      endif
    enddo

    do
      call scale_factor(a,da,dt_e)
      if (da.gt.DA_MAX) then
        dt_e=dt_e*(DA_MAX/da)
      else
        exit
      endif
    enddo

    if (nts.eq.1) dt_acc=100.0

    dt=min(dt_acc,dt_e)
    call scale_factor(a,da,dt)

    if (a+da.gt.a_cp(cur_cp)) then
      check_step=.true.
      dt=dt*(a_cp(cur_cp)-a)/da
      call scale_factor(a,da,dt)
      if (cur_cp.eq.num_cp) stopsim=.true.
    endif

    a=a+da
    t=t+dt
    aGdt=(a+da/2)/(6*acos(-1.0))*dt
    aGdtC=0.0
    dtC=0.0

    write(*,*) 'dt_old=',dt_old
    write(*,*) 'dt=',dt
    write(*,*) 'fine acceleration dt=',dt_acc
    write(*,*) 'expansion dt=',dt_e
    write(*,*) 'a=',a
    write(*,*) 'z=',1.0/a-1.0
    write(*,*) 't=',t
    write(*,*) 'aGdt=',aGdt

  else

! update the timestep number within the current timestep
! sweep

    ntsw=ntsw+1
    if (ntsw > ts_ratio) ntsw = 1

! if we are at the middle step of the sweep, then we 
! calculate the coarse mesh velocity update

    if (ntsw.eq.ts_ratio/2+1) then
      coarse_step=.true.
    else
      coarse_step=.false.
    endif

! if we are at the first step of the sweep, then we calculate
! the timestep parameters for the entire sweep since dt must
! be fixed for the entire sweep.

    if (ntsw.eq.1) then

      check_sweep=.false.
      stop_sweep=.false.
      proj_sweep=.false.

      dt_e=0.5
      do
        call scale_factor(a,da,dt_e)
        ra=da/(a+da)
        if (ra.gt.RA_MAX) then
          dt_e=dt_e*(RA_MAX/ra)
        else
          exit
        endif
      enddo
      do
        call scale_factor(a,da,dt_e)
        if (da.gt.DA_MAX) then
          dt_e=dt_e*(DA_MAX/da)
        else
          exit
        endif
      enddo  

! if we are at the first timestep, then we set the timestep ratio to 1,
! otherwise it needs to be calculated for the sweep

      if (nts.eq.1) then
        dt_acc=100.0
        dt_acc_C=100.0
        ts_ratio=1
      else
        if (dt_acc < dt_acc_C) then
          if (dt_e < dt_acc) then
            ts_ratio=1
          else
            ts_ratio=int((dt_e/dt_acc)/2)*2+1 
          endif
        else
          if (dt_e < dt_acc_C) then
            ts_ratio=1
          else
            ts_ratio=int((dt_e/dt_acc_C)/2)*2+1
          endif
        endif
        if (ts_ratio > TS_RATIO_MAX) ts_ratio = TS_RATIO_MAX
      endif

! now we have the values for dt and the timestep ratio, need to select
! value of dt used in sweep
      
      write(*,*) 'Timestep',nts
      write(*,*) 'dt_e=',dt_e,'dt_fine=',dt_acc,'dt_coarse=',dt_acc_C 

      dt=min(dt_acc,dt_acc_C,dt_e)
      dt=dt*DT_SWEEP_SCALE 
      astemp=a
      do i=1,ts_ratio
        call scale_factor(astemp,das(i),dt)      
        astemp=astemp+das(i)
      enddo

      if (a_cp(cur_cp).lt.a_rp(cur_rp).or.cur_rp.gt.num_rp) then
        if (astemp.gt.a_cp(cur_cp)) then
          check_sweep=.true.
          dt=dt*(a_cp(cur_cp)-a)/sum(das)
          write(*,*) 'das=',das
          astemp=a
          do i=1,ts_ratio
            call scale_factor(astemp,das(i),dt)
            astemp=astemp+das(i)
          enddo
          if (cur_cp.eq.num_cp) stop_sweep=.true.
        endif
      else
        if (astemp.gt.a_rp(cur_rp)) then
          proj_sweep=.true.
          dt=dt*(a_rp(cur_rp)-a)/sum(das)
          astemp=a
          do i=1,ts_ratio
            call scale_factor(astemp,das(i),dt)
            astemp=astemp+das(i)
          enddo
        endif
      endif

      dtC=dt*ts_ratio

    endif

    da=das(ntsw)
    aGdt=(a+da/2)/(6*acos(-1.0))*dt
    aGdtC=(a+da/2)/(6*acos(-1.0))*dtC
    if (ntsw.eq.ts_ratio.and.check_sweep) check_step=.true.
    if (ntsw.eq.ts_ratio.and.proj_sweep) proj_step=.true.
    if (ntsw.eq.ts_ratio.and.stop_sweep) stopsim=.true.

    write(*,*) 'Timestep',nts,'in sweep:',ntsw
    write(*,*) 'dt=',dt
    write(*,*) 'ai=',a,'af=',a+da
    write(*,*) 'zi=',1.0/a-1.0,'zf=',1.0/(a+da)-1.0
    write(*,*) 'ti=',t,'tf=',t+dt
    a=a+da
    t=t+dt
    write(*,*) 'aGdt=',aGdt
    write(*,*) 'aGdtC=',aGdtC

  endif

endif

! broadcast timestep parameters

call mpi_bcast(a,1,mpi_real,0,mpi_comm_world,ierr)
call mpi_bcast(check_step,1,mpi_logical,0,mpi_comm_world,ierr)
call mpi_bcast(stopsim,1,mpi_logical,0,mpi_comm_world,ierr)
call mpi_bcast(proj_step,1,mpi_logical,0,mpi_comm_world,ierr)
call mpi_bcast(coarse_step,1,mpi_logical,0,mpi_comm_world,ierr)
call mpi_bcast(aGdt,1,mpi_real,0,mpi_comm_world,ierr)
call mpi_bcast(dt,1,mpi_real,0,mpi_comm_world,ierr)
call mpi_bcast(aGdtC,1,mpi_real,0,mpi_comm_world,ierr)
call mpi_bcast(dtC,1,mpi_real,0,mpi_comm_world,ierr)

tsfile=TSPATH//'timestep.log'
open(unit=73,file=tsfile,position='append')
write(73,'(i4,1x,i2,1x,f6.4,1x,f6.2,1x,f6.4,1x,f6.4,1x,f8.4,1x,f8.4)') &
           nts,ntsw,a,t,dt,dt_e,dt_acc,dt_acc_C
close(73)

end subroutine multistep

!*************************************************************
! scale_factor subroutine - calculate da given a and dt

subroutine scale_factor(a,da,dt)
    implicit none
    include 'cosmopar.fh'

    real(kind=4), intent(in) :: a,dt
    real(kind=4), intent(out) :: da
    real (kind=4) :: adot,addot,atdot,arkm,a3rlm,omHsq

    !! Expand Friedman equation to third order and integrate
    omHsq=4./9
    a3rlm=a**3*OMEGA_L/OMEGA_M
    arkm=a*(1-OMEGA_M-OMEGA_L)/OMEGA_M
    adot=sqrt(omHsq*a**3*(1+arkm+a3rlm))
    addot=a**2*omHsq*(1.5+2*arkm+3*a3rlm)
    atdot=a*adot*omHsq*(3+6*arkm+15*a3rlm)
    da=adot*dt+(addot*dt**2)/2+(atdot*dt**3)/6

end subroutine scale_factor

